/** This file is part of Reactor.
 *  Copyright (c) 2020-2022 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */

export const version = 22300;

import api from '/client/ClientAPI.js';
import Rulesets from "/client/Rulesets.js";
import Ruleset from "/client/Ruleset.js";
import Rule from "/client/Rule.js";

import * as Common from './reactor-ui-common.js';

import { _T, getLocale } from './i18n.js';

import { RuleEditor, condTypeName, weekdayNames } from "./rule-editor.js";
import ReactionEditor from "./reaction-editor.js";
import ExpressionEditor from "./expression-editor.js";

/* Frequently used */
var isEmpty = Common.isEmpty;
var idSelector = Common.idSelector;
var showSysModal = Common.showSysModal;
var shortTime = Common.shortTime;

import Tab from "./tab.js";

export default (function($) {

    class RulesTab extends Tab {
        constructor( $parent ) {
            super( 'tab-rules', $parent );

            this.configModified = false;
            this.editors = {};
            this.editrule = false;
            this.newrule = false;

            let self = this;
            $( 'li#nav-item-rules div.nav-item-controls' ).hide();
            $( 'li#nav-item-rules div.nav-item-controls a#nav-addruleset' ).on( 'click.reactor', async function() {
                var ruleset = await Ruleset.create();
                var $nav = $( '<li class="nav-item re-nav-sub w-100"></li>' )
                    .prependTo( 'ul#ruleset-nav' );
                $( '<a href="#" class="nav-link"></a>' )
                    .attr( 'id', ruleset.id )
                    .text( ruleset.name || ruleset.id )
                    .on( 'click.reactor', self.handleRulesetClick.bind( self ) )
                    .appendTo( $nav );
                $( 'a.nav-link', $nav ).triggerHandler( 'click.reactor' );
            });
        }

        activate() {
            this.show_list();
        }

        canSuspend() {
            /* If editing, can't suspend. */
            if ( "edit" === this.$tab.data( 'mode' ) && this.configModified ) {     /* ??? this.editRule not false? */
                console.log("Stopping activate event in edit mode");
                return false;
            }
            return true;
        }

        suspending() {
            var $subnav = $( 'li#nav-item-rules ul#ruleset-nav' );
            $subnav.removeClass( "show" ).addClass( "hide" ).slideUp();
        }

        getEditor( id ) {
            return this.editors[ id ];
        }

        /**
         * Update save/revert buttons
         */
        updateSaveControls() {
            var pos = $( window ).scrollTop();
            var cansave = this.configModified;
            // console.log("updateSaveControls() configModified", this.configModified);
            var self = this;
            Object.keys( this.editors ).forEach( function(k) {
                if ( ! self.editors[k].canSave() ) {
                    // console.log("updateSaveControls() editor",k,"reports cannot save");
                    cansave = false;
                }
            });
            $('button.saveconf').prop('disabled', ! cansave )
                .attr( 'title', this.configModified ?
                    ( cansave ? "" : _T("Fix errors before saving")) : _T("No unsaved changes") );
            $('button.revertconf').prop('disabled', false )
                .attr( 'title', this.configModified ?
                    _T("You have unsaved changes!") : _T("No unsaved changes") );
            setTimeout( function() { $(window).scrollTop( pos ); }, 100 );
        }

        async editRule( rule, isNew ) {
            var self = this;

            this.editrule = rule.getDataObject().value;
            this.newrule = !!isNew;

            this.$tab.empty();
            this.$tab.data('mode', 'edit').attr( 'data-mode', 'edit' );
            this.$tab.data('rule', rule.id).attr( 'data-rule', rule.id);
            // this.$tab.data('ruleset', rule.__set.id).attr( 'data-ruleset', rule.__set.id );

            var $container = $('<div id="re-edit-head" class="container-fluid"></div>' )
                .appendTo( this.$tab );
            var $row = $( '<div class="row"></div>' ).appendTo( $container );
            $( '<div class="col"><h1></h1></div>' ).appendTo( $row );
            $( `<div class="col text-end mt-3">
  <button class="btn btn-sm btn-success saveconf">${_T(['#edit-button-save','Save'])}</button>
  <button class="btn btn-sm btn-danger revertconf">${_T(['#edit-button-exit','Exit'])}</button>
</div>` ).appendTo( $row );
            $( 'h1', $row ).text( rule.name || rule.id ).on( 'click.reactor', function( ev ) {
                var $el = $( ev.currentTarget );
                var $form = $( '<div class="form-group"><input id="newname" class="form-control form-control-sm"></div>' );
                $( 'input', $form ).val( rule.name || rule.id );
                showSysModal({
                    title: _T(['#rule-rename-title','Rename Rule']),
                    body: $form,
                    buttons: [{
                        class: "btn-primary",
                        label: _T(['#rule-rename-button-cancel','Cancel']),
                        close: true
                    },{
                        class: "btn-success",
                        event: "rename",
                        label: _T(['#rule-rename-button-rename','Rename'])
                    }
                ]
                }).then( data => {
                    if ( "rename" === data ) {
                        var newname = $( '#sysdialog #newname' ).val() || "";
                        if ( "" !== newname && newname !== self.editrule.name ) {
                            self.editrule.name = newname;
                            $el.text( newname );
                            self.getEditor( 'trig_editor' ).signalModified();
                        }
                    }
                });
            });

            $container = $( '<div id="ct-trig"></div>' ).appendTo( this.$tab );
            $row = $( '<div class="row re-sectionhead no-gutters align-items-center"></div>' ).appendTo( $container );
            $( `<div class="col">
  <h6>
    <a class="bi-btn bi-chevron-down mx-2" data-bs-toggle="collapse" data-bs-target="#trig-editor"
         role="button" aria-expanded="true" aria-controls="trig-editor"></a>
    ${_T(['#ruleedit-triggers-title','Triggers'])}
  </h6>
</div>` )
                .appendTo( $row );
            if ( ! ( rule.triggers && "group" === rule.triggers.type ) ) {
                rule.triggers = { type: "group", op: "and", conditions: [], id: rule.id + '-trig',
                    disabled: true };
            } else {
                /* ??? Temporary fix adopted from importer. Root trigger group can't have options, so for
                   users that have already imported and started work, if their triggers got imported with
                   options, push the triggers down to a new subgroup. */
                /* If root (trig) has options, no can do, push down to new group with new ID */
                if ( Object.keys( rule.triggers.options || {} ).length > 0 ) {
                    let new_grp = { ...rule.triggers }; /* copy */
                    rule.triggers.conditions = [ { id: Common.getUID( 'cond' ), type: "comment", data: { comment: "Rule repaired; delete this comment and save the updated rule." } }, new_grp ];
                    delete rule.triggers.options;
                    new_grp.id = Common.getUID( 'grp' );
                    new_grp.name = 'Wrapper group for options';
                    delete new_grp.disabled;
                    delete new_grp.options.after;
                    delete new_grp.options.aftertime;
                    delete new_grp.options.aftermode;
                }
            }
            var trig_editor = new RuleEditor( rule.triggers || {}, $container, { fixedName: '', contextRule: rule } );
            this.editors.trig_editor = trig_editor;
            trig_editor.editor()
                .attr( 'id', 'trig-editor' )
                .addClass( "collapse show" )
                .on( 'update-controls', function( ev, ui ) {
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'modified', function( ev, ui ) {
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'ready', function( ev, ui ) {
                    ui.editor.editor().collapse( "show" );
                });

            $container = $( '<div id="ct-cons"></div>' ).appendTo( this.$tab );
            $row = $( '<div class="row re-sectionhead no-gutters align-items-center"></div>' ).appendTo( $container );
            $( `<div class="col">
  <h6>
    <a class="bi-btn bi-chevron-down mx-2" data-bs-toggle="collapse" data-bs-target="#const-editor"
       role="button" aria-expanded="true" aria-controls="const-editor"></a>
    ${_T(['#ruleedit-constraints-title','Constraints'])}
  </h6>
</div>` )
                .appendTo( $row );
            /* ??? temporary remove later */
            if ( rule.conditions ) {
                var d = rule.getDataObject();
                d.value.constraints = d.value.conditions;
                delete d.value.conditions;
                d.save();
            }
            /* ??? end temporary */
            if ( ! ( rule.constraints && "group" === rule.constraints.type ) ) {
                rule.constraints = { type: "group", op: "and", conditions: [], id: 'cons' };
            }
            delete rule.constraints.options; /* Constraints can never have options */
            var const_editor = new RuleEditor( rule.constraints || {}, $container,
                { fixedName: '', contextRule: rule, constraints: true } );
            this.editors.const_editor = const_editor;
            const_editor.editor()
                .attr( 'id', 'const-editor' )
                .addClass( "collapse show" )
                .on( 'update-controls', function( ev, ui ) {
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'modified', function( ev, ui ) {
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'ready', function( ev, ui ) {
                    ui.editor.editor().collapse( ((rule.constraints || {}).conditions || []).length > 0 ? "show" : "hide" );
                });

            $container = $( '<div id="ct-set"></div>' ).appendTo( this.$tab );
            $row = $( '<div class="row re-sectionhead no-gutters align-items-center"></div>' ).appendTo( $container );
            $( `<div class="col">
  <h6>
    <a class="bi-btn bi-chevron-down mx-2" data-bs-toggle="collapse" data-bs-target="#set-react-editor"
       role="button" aria-expanded="true" aria-controls="set-react-editor"></a>
    ${_T(['#ruleedit-setreaction-title','Set Reaction'])}
  </h6>
</div>` )
                .appendTo( $row );
            var reaction = rule.react_set;
            if ( !reaction ) {
                rule.react_set = reaction = { id: rule.id + ':S', rule: rule.id, actions: [] };
                this.configModified = true;
            }
            delete reaction.name;
            var set_editor = new ReactionEditor( reaction, $container, { fixedName: "", contextRule: rule } );
            this.editors.set_editor = set_editor;
            set_editor.editor()
                .attr( 'id', 'set-react-editor' )
                .addClass( "collapse show" )
                .on( 'update-controls', function( ev, ui ) {
                    /* During editing, various changes make the reaction saveable, or not. */
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'modified', function( ev, ui ) {
                    /* Sent when the cached reaction is modified (no errors) */
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'ready', function( ev, ui ) {
                    ui.editor.editor().collapse( "show" );
                });

            $container = $( '<div id="ct-reset"></div>' ).appendTo( this.$tab );
            $row = $( '<div class="row re-sectionhead no-gutters align-items-center"></div>' ).appendTo( $container );
            $( `<div class="col">
  <h6>
    <a class="bi-btn bi-chevron-down mx-2" data-bs-toggle="collapse" data-bs-target="#reset-react-editor"
       role="button" aria-expanded="true" aria-controls="reset-react-editor"></a>
    ${_T(['#ruleedit-resetreaction-title','Reset Reaction'])}
  </h6>
</div>` )
                .appendTo( $row );
            reaction = rule.react_reset;
            if ( !reaction ) {
                rule.react_reset = reaction = { id: rule.id + ':R', rule: rule.id, actions: [] };
                this.configModified = true;
            }
            delete reaction.name;
            var reset_editor = new ReactionEditor( reaction, $container, { fixedName: "", contextRule: rule } );
            this.editors.reset_editor = reset_editor;
            reset_editor.editor()
                .attr( 'id', 'reset-react-editor' )
                .addClass( "collapse show" )
                .on( 'update-controls', function( ev, ui ) {
                    /* During editing, various changes make the reaction saveable, or not. */
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'modified', function( ev, ui ) {
                    /* Sent when the cached reaction is modified (no errors) */
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'ready', function( ev, ui ) {
                    ui.editor.editor().collapse( (reaction.actions || []).length > 0 ? "show" : "hide" );
                });

            /* Expressions editor */
            $container = $( '<div id="ct-expr"></div>' ).appendTo( this.$tab );
            $row = $( '<div class="row re-sectionhead no-gutters align-items-center"></div>' ).appendTo( $container );
            $( `<div class="col">
  <h6>
    <a class="bi-btn bi-chevron-down mx-2" data-bs-toggle="collapse" data-bs-target="#expr-editor"
       role="button" aria-expanded="true" aria-controls="expr-editor"></a>
    ${_T(['#ruleedit-expressions-title','Local Expressions'])}
  </h6>
</div>` )
                .appendTo( $row );
            var expressions = rule.expressions || {};
            var expr_editor = new ExpressionEditor( expressions, $container, { isGlobal: false, contextRule: rule } );
            this.editors.expr_editor = expr_editor;
            expr_editor.editor()
                .attr( 'id', 'expr-editor' )
                .addClass( "collapse show" )
                .on( 'update-controls', function( ev, ui ) {
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'modified', function( ev, ui ) {
                    /* Sent when data modified */
                    self.configModified = self.configModified || ui.modified;
                    self.updateSaveControls();
                }).on( 'ready', function( ev, ui ) {
                    ui.editor.editor().collapse( Common.hasAnyProperty( expressions ) ? "show" : "hide" );
                });

            /* Show/hide collapsibles */
            $( '.collapse', this.$tab ).on( 'show.bs.collapse', function( ev ) {
                var $el = $( ev.currentTarget );
                var id = $el.attr( 'id' );
                $( 'a[data-bs-toggle="collapse"][data-bs-target="#' + idSelector( id ) + '"]', this.$tab )
                    .removeClass( 'bi-chevron-up' ).addClass( 'bi-chevron-down' );
                /* Remove hidden content warning badge (see below) */
                const $ee = $( 'div.re-embeddable-editor#' + idSelector( id ) );
                const $h = $ee.parent().children( 'div.re-sectionhead' );
                $h.find( 'span.badge' ).remove();
            }).on( 'hidden.bs.collapse', function( ev ) {
                var $el = $( ev.currentTarget );
                var id = $el.attr( 'id' );
                $( 'a[data-bs-toggle="collapse"][data-bs-target="#' + idSelector( id ) + '"]', this.$tab )
                    .removeClass( 'bi-chevron-down' ).addClass( 'bi-chevron-up' );
                /* Place hidden content warning badge if editor is not empty. */
                const $ee = $( 'div.re-embeddable-editor#' + idSelector( id ) );
                const $h = $ee.parent().children( 'div.re-sectionhead' );
                if ( $ee.data( "embeddable-editor" ).isEmpty() ) {
                    $h.find( 'span.badge' ).remove();
                } else {
                    $( '<span class="badge bg-info fw-normal">!</span>' )
                        .attr( 'title', _T('Content hidden') )
                        .appendTo( $( 'div.col h6', $h ) );
                }
            });

            trig_editor.edit();
            const_editor.edit();
            set_editor.edit();
            reset_editor.edit();
            expr_editor.edit();

            $( 'button.saveconf,button.revertconf', this.$tab )
                .on( 'click.reactor', this.handleSaveRevertClick.bind( this ) );

            this.configModified = this.newrule;
            this.updateSaveControls();
        }

        handleRuleDrop( ev, ui ) {
            var $dragged = $( ui.draggable );
            var $droppedon = $( ev.target );
            console.log($dragged, $droppedon);
            var dragid = $dragged.attr( 'id' ).replace( /^rule-/, "" );
            var dropid = $droppedon.find( 'a.nav-link' ).attr( 'id' );
            var source_ruleset_id = $( 'div#rules' ).data( 'ruleset' );
            if ( dropid === source_ruleset_id ) {
                /* No change */
                return;
            }
            var source_ruleset = Ruleset.getInstance( source_ruleset_id );
            var ix = source_ruleset.getRuleIDs().indexOf( dragid );
            if ( ix >= 0 ) {
                var dest_ruleset = Ruleset.getInstance( dropid );
                var dobj = dest_ruleset.getDataObject();
                /* Modify Data objects in place. Add to dest first. */
                dobj.value.rules.push( dragid );
                dobj.save().then( () => {
                    var dobj = source_ruleset.getDataObject();
                    dobj.value.rules.splice( ix, 1 );
                    dobj.save();
                    $( 'div#rightheader button#deleteruleset' )
                        .prop( 'disabled', source_ruleset.value.rules.length > 0 );
                });
            }
        }

        async handleRuleSort( ev, ui ) {
            var $container = $( 'div#rules' );
            var ruleset_id = $container.data( 'ruleset' );
            var ruleset = await Ruleset.getInstance( ruleset_id );
            var $items = $( '.re-rule', $container );
            var rsdata = ruleset.getDataObject();
            rsdata.value.rules.length = 0; /* clear array in place */
            $items.each( ( ix, obj ) => {
                var id = $( obj ).attr( 'id' ).replace( /^rule-/, "" );
                rsdata.value.rules.push( id );
            });
            rsdata.save();
        }

        /**
         *  Called when a (drag-sorted) ruleset is dropped in some position.
         *  Reorder the rulesets to the displayed order.
         */
        async handleRulesetSort( ev, ui ) {
            var rulesets = await Rulesets.getInstance();
            var rsdata = rulesets.getDataObject();
            /* The ruleset's DO value is just an array of ruleset IDs */
            rsdata.value.length = 0; /* clear in place */
            var $items = $( 'ul#ruleset-nav li.nav-item' );
            $items.each( ( ix, obj ) => {
                var id = $('a.nav-link', obj).attr( 'id' );
                rsdata.value.push( id );
            });
            rsdata.save();
        }

        async showRuleStateData( ruleid, $row, cstate, initial ) {
            let rule = await Rule.getInstance( ruleid );
            var $st = $( 'div.re-rulestatus', $row );
            let last = $st.data( 'laststate' );
            let state = (cstate.rule || {}).evalstate;
            $st.data( 'laststate', state );
            let changed = last !== state;
            if ( "undefined" === typeof state ) {
                $st.text( '-' );
            } else if ( null === state ) {
                $st.text( _T(['#rule-state-null','(none)']) );
            } else {
                $st.text(
                    ( ( cstate.rule.evalstate ? _T(['#rule-state-true','SET']) : _T(['#rule-state-false','reset']) ) ) +
                    ' ' +
                    _T( 'since {0}', shortTime( cstate.rule.evalstamp ), cstate.rule.evalstamp )
                );
            }
            if ( changed && !initial ) {
                $row.stop().clearQueue().css( 'background-color', '#0c0' );
                $row.animate( { backgroundColor: '#fff' }, 2000);
            }

            $row.toggleClass( 'trouble', !!( cstate.rule || {} ).trouble );
            if ( ( cstate.rule || {} ).trouble ) {
                if ( 0 === $( '.col.re-rulename i.trouble-icon', $row ).length ) {
                    $( '.col.re-rulename', $row )
                        .prepend( '<i class="bi bi-exclamation-triangle-fill trouble-icon me-1"></i>' );
                }
            } else {
                $( '.col.re-rulename i.trouble-icon', $row ).remove();
            }
            Object.keys( cstate ).forEach( condid => {
                var st = cstate[ condid ] || {};
                var $line = $( '.card div#' + idSelector( ruleid + "-cond-" + condid ), $row );
                var $el = $( 'span.currentvalue', $line );
                let evt = null;
                if ( st ) {
                    evt = st.lastvalue;
                }
                /* Remapping/humanizing of values */
                switch ( $el.closest( 'div' ).data( 'cond-type' ) ) {

                    case 'comment':
                        $( 'div.condind i', $line ).hide();
                        $el.text( "" );
                        return; /* display nothing */

                    case 'weekday':
                        /* Weekday stores a numeric day of the week */
                        if ( evt ) {
                            evt = weekdayNames[ evt ] || evt;
                        }
                        break;

                    case 'sun':
                    case 'trange':
                    case 'interval':
                        /* Timestamps */
                        if ( evt ) {
                            try {
                                evt = shortTime( evt );
                            } catch ( err ) {
                                console.log( "Invalid timestamp", evt );
                            }
                        }
                        break;

                    default:
                        /* Nada */
                }
                if ( "boolean" === typeof evt ) {
                    evt = _T( String( evt ) );
                } else if ( null !== evt && "object" === typeof evt ) {
                    try {
                        evt = JSON.stringify( evt );
                    } catch ( err ) {
                        console.error( "Failed to stringify():", err, evt );
                    }
                } else if ( "string" !== typeof evt ) {
                    evt = String( evt );
                }
                let est = _T( '{0} -- {1} as of {2}', evt, _T( String( Common.coalesce( st.evalstate ) ) ),
                    shortTime( st.evalstamp ), st.evalstamp );
                $el.text( est )
                    .attr( 'title', String( evt ) );
                $line.toggleClass( 'truestate', true === st.evalstate )
                    .toggleClass( 'falsestate', false === st.evalstate )
                    .toggleClass( 'nullstate', null === st.evalstate || "boolean" !== typeof st.evalstate );
                $( 'div.condind i', $line ).toggleClass( 'bi-check-square text-success', true === st.evalstate )
                    .toggleClass( 'bi-x-square text-danger', false === st.evalstate )
                    .toggleClass( 'bi-square text-muted', null === st.evalstate || "boolean" !== typeof st.evalstate );

                if ( "undefined" !== typeof st.repeats && st.repeats.length > 1 ) {
                    var dtime = Math.floor( ( st.repeats[ st.repeats.length - 1 ] - st.repeats[0] ) / 100 ) / 10;
                    $el.append( `<span ${_T(['#cond-status-repeats','(last {0} span {1:.1f} seconds)'], st.repeats.length, dtime)}</span>` );
                }
                if ( ( st.pulsecount || 0 ) > 0 ) {
                    var lim = 0; // ( cond.options||{} ).pulsecount || 0;
                    $el.append( `<span> ${_T(['#cond-status-pulsed','(pulsed {0:d?<999}{0:"&gt;999"?!<999}{1:" of max "?>0}{1:d?>0} times)'],
                        st.pulsecount, lim)}</span>` );
                }
                if ( st.latched ) {
                    $el.append( `<span> ${_T(['#cond-status-latched','(latched)'])}</span>` );
                }

                /* Generate unique IDs for timers so that redraws will have
                   different IDs, and the old timers will self-terminate. */
                var self = this;
                if ( rule.disabled ) {
                    $( 'span.timer', $el ).remove();
                    $line.removeClass( 'reactor-timing' ).stop( true, true );
                } else if ( st.laststate && st.waituntil ) {
                    let id = Common.getUID( 'timer-' );
                    $line.addClass('reactor-timing');
                    $('<span class="timer"></span>').attr( 'id', id ).appendTo( $el );
                    (function( c, t, l ) {
                        setTimeout( function() { self.updateTime( c, t, _T(['#cond-status-sustained','sustained']), false, l ); }, 20 );
                    })( id, st.statestamp, st.duration );
                } else if ( st.evalstate && st.holduntil ) {
                    let id = Common.getUID( 'timer-' );
                    $line.addClass('reactor-timing');
                    $('<span class="timer"></span>').attr( 'id', id ).appendTo( $el );
                    (function( c, t, l ) {
                        setTimeout( function() { self.updateTime( c, t, _T(['#cond-status-resetdelay','reset delay']), true, l ); }, 20 );
                    })( id, st.holduntil, 0 );
                } else if ( st.pulsenext ) {
                    let id = Common.getUID( 'timer-' );
                    $line.addClass('reactor-timing');
                    $('<span class="timer"></span>').attr( 'id', id ).appendTo( $el );
                    (function( c, t, l ) {
                        setTimeout( function() { self.updateTime( c, t, _T(['#cond-status-pulsebreak','break']), true, l ); }, 20 );
                    })( id, st.pulsenext, 0 );
                } else if ( st.pulseuntil ) {
                    let id = Common.getUID( 'timer-' );
                    $line.addClass('reactor-timing');
                    $('<span class="timer"></span>').attr( 'id', id ).appendTo( $el );
                    (function( c, t, l ) {
                        setTimeout( function() { self.updateTime( c, t, _T(['#cond-status-pulsing','pulse']), true, l ); }, 20 );
                    })( id, st.pulseuntil, 0 );
                } else {
                    $( 'span.timer', $el ).remove();
                    $line.removeClass( 'reactor-timing' ).stop( true, true );
                }

                if ( st.next_edge ) {
                    $( `<span>; ${_T(['#cond-status-waiting-time','waiting for {0}'], shortTime( st.next_edge ), st.next_edge)}</span>` )
                        .appendTo( $el );
                }
            });

            $( 'div.expr-status div.expr-status-expr', $row ).each( ( ix, obj ) => {
                let $ex = $( obj );
                let name = $ex.data( 'name' ) || "";
                if ( ( cstate.expr || {})[ name ] ) {
                    let st = cstate.expr[ name ] || {};
                    if ( ( st.laststamp || 0) !== ( $ex.data( 'since' ) || 0 ) ) {
                        let val = "";
                        let stamp = "-";
                        val = "(" + typeof( st.lastvalue ) + ") ";
                        try {
                            val += JSON.stringify( st.lastvalue );
                        } catch ( err ) {
                            val += String( st.lastvalue );
                        }
                        if ( ( st.laststamp || 0 ) > 0 ) {
                            stamp = _T( 'since {0}', shortTime( st.laststamp ), st.laststamp );
                        }
                        $( 'div.expr-value', $ex ).text( val );
                        $( 'div.expr-stamp', $ex ).text( stamp );
                        $ex.data( 'since', st.laststamp || 0 );
                        $ex.stop().clearQueue().css( 'background-color', '#0c0' );
                        $ex.animate( { backgroundColor: '#fff' }, 2000);
                    }
                } else {
                    $( 'div.expr-value', $ex ).text( _T(['#expr-status-no-data','(no data)']) );
                    $( 'div.expr-stamp', $ex ).text( "-" );
                }
            });
        }

        async handleDataObjectChange( msg ) {
            if ( "data-changed" === msg.type ) {
                var dobj = msg.data;
                var ruleid = dobj.id.replace( /^cs-/, "" );
                var $row = $( 'div#rules div#' + idSelector( 'rule-' + ruleid ) + '.re-rule' );
                this.showRuleStateData( ruleid, $row, dobj.value );
            } else if ( "data-deleted" === msg.type ) {
                console.log("**** DELETED DATA ****",msg);
            }
        }

        async handleRuleEnableToggle( ev ) {
            var $el = $( ev.currentTarget );
            var $row = $el.closest( '.re-rule' );
            var id = $row.attr( 'id' ).replace( /^rule-/, "" );
            var rule = await Rule.getInstance( id );
            api.enableRule( id, !!rule.disabled ).then( data => {
                var newstate = !!data.value.triggers.disabled;
                $row.toggleClass( 'rule-disabled', newstate );
                $( 'button.ruleEnable i', $row )
                    .toggleClass( 'bi-toggle-off', newstate )
                    .toggleClass( 'bi-toggle-on', !newstate );
            });
        }

        async handleRuleReset( ev ) {
            var $el = $( ev.currentTarget );
            var $row = $el.closest( '.re-rule' );
            var id = $row.attr( 'id' ).replace( /^rule-/, "" );
            $el.addClass( "active" );
            var rule = await Rule.getInstance( id );
            await rule.reset().finally( () => {
                $el.removeClass( "active" );
            });
            return false;
        }

        async handleRuleCopy( ev ) {
            var $el = $( ev.currentTarget );
            var $row = $el.closest( '.re-rule' );
            var id = $row.attr( 'id' ).replace( /^rule-/, "" );
            var rule = await Rule.getInstance( id );
            var rulesets = await Rulesets.getRulesets();
            var $mm = $( '<select id="rulesetlist" class="form-select form-select-sm ms-1"></select>' );
            $( '<option></option>' ).val( "" ).text( _T('(to current rule set)') ).appendTo( $mm );
            var current_ruleset = false;
            rulesets.forEach( set => {
                if ( null === set ) { /* unable to fetch set in sequence */
                    return;
                }
                if ( set.getRuleIDs().indexOf( rule.getID() ) >= 0 ) {
                    current_ruleset = set;
                    return;
                }
                $( '<option></option>' ).val( set.id ).text( set.name || set.id )
                    .appendTo( $mm );
            });
            var $div = $( '<form></form>' );
            $( `<label for="rulesetlist">${_T(['#rule-move-target-label','Destination Rule Set:'])}</label>` )
                .appendTo( $div );
            $mm.appendTo( $div );
            showSysModal( {
                title: _T(['#rule-move-title','Copy/Move Rule']),
                body: $div,
                buttons: [{
                    label: _T(['#rule-move-button-move','Move']),
                    event: "move",
                    class: "btn-warning"
                },{
                    label: _T(['#rule-move-button-copy','Copy']),
                    event: "copy",
                    class: "btn-success"
                },{
                    label: _T(['#rule-move-button-cancel','Cancel']),
                    close: true,
                    class: "btn-primary"
                }]
            }).then( async event => {
                let rsid = $( '#sysdialog select#rulesetlist' ).val() || "";
                if ( "move" === event ) {
                    if ( isEmpty( rsid ) || rsid === current_ruleset.id ) {
                        /* Already in dest ruleset */
                        return event;
                    }
                    let dest_ruleset = await Ruleset.getInstance( rsid );
                    await rule.move( dest_ruleset );
                    $row.remove();
                } else if ( "copy" === event ) {
                    /* Synthesize a new rule with a new ID in target ruleset */
                    let dest_ruleset;
                    if ( isEmpty( rsid ) ) {
                        dest_ruleset = current_ruleset;
                    } else {
                        dest_ruleset = await Ruleset.getInstance( rsid );
                    }
                    await rule.copy( dest_ruleset );
                    if ( isEmpty( rsid ) ) {
                        /* Repaint current ruleset to show new rule here. */
                        $( 'ul#ruleset-nav li.nav-item a#' + idSelector( current_ruleset.id ) ).click();
                    }
                } else {
                    /* Cancel */
                }
            });
            return false;
        }

        async handleRuleDelete( ev ) {
            var $el = $( ev.currentTarget );
            var $row = $el.closest( '.re-rule' );
            var id = $row.attr( 'id' ).replace( /^rule-/, "" );
            try {
                var rule = await Rule.getInstance( id );
                showSysModal( {
                    title: _T(['#rule-delete-title','Delete Rule']),
                    body: _T(['#rule-delete-prompt','Really delete {0:q}?'], rule.name || id),
                    buttons: [{
                        label: _T(['#rule-delete-button-delete','Delete']),
                        event: "delete",
                        class: "btn-danger"
                    },{
                        label: _T(['#rule-delete-button-cancel','Cancel']),
                        close: true,
                        class: "btn-success"
                    }]
                }).then( event => {
                    if ( "delete" === event ) {
                        rule.delete().then( () => {
                            $row.remove();
                            $( 'div#rightheader button#deleteruleset' )
                                .prop( 'disabled', $( 'div#rules .re-rule' ).length > 0 );
                        }).catch( err => {
                            console.err( err );
                            showSysModal( {
                                title: _T(['#dialog-error-title','Error']),
                                body: _T(['#rule-delete-error','An error occurred while deleting the rule.'])
                            } );
                        });
                    }
                });
            } catch ( err ) {
                console.error( err );
                showSysModal( {
                    title: _T(['#dialog-error-title','Error']),
                    body: _T('The operation cannot be completed at this time.')
                } );
            }
            return false;
        }

        updateTime( elid, target, prefix, countdown, limit ) {
            var $el = $( '#' + idSelector( elid ) );
            if ( 0 === $el.length ) {
                return; /* quietly die */
            }
            var now = Date.now();
            var d;
            if ( countdown ) {
                /* Count down -- delta is (future) target to now */
                d = target - now;
                if ( d < ( limit || 0 ) ) {
                    let $line = $el.closest( ".cond" );
                    $el.remove();
                    if ( 0 === $( 'span.timer', $line ).length ) {
                        // Stop animation if no timers remain
                        $line.removeClass( 'reactor-timing' ).stop( true, true );
                    }
                    return;
                }
            } else {
                /* Count up -- delta is now since target */
                d = now - target;
                if ( limit && d > limit ) {
                    let $line = $el.closest( ".cond" );
                    $el.remove();
                    if ( 0 === $( 'span.timer', $line ).length ) {
                        // Stop animation if no timers remain
                        $line.removeClass( 'reactor-timing' ).stop( true, true );
                    }
                    return;
                }
            }
            d = Math.floor( d / 1000 + 0.5 );
            var hh = Math.floor( d / 3600 );
            d -= hh * 3600;
            var mm = Math.floor( d / 60 );
            d -= mm * 60;
            d = (mm < 10 ? '0' : '') + String(mm) + ':' + (d < 10 ? '0' : '') + String(d);
            if ( 0 !== hh ) {
                d = (hh < 10 ? '0' : '') + String(hh) + ':' + d;
            }
            $el.text( `; ${prefix} ${d}` );
            var self = this;
            setTimeout( function() { self.updateTime( elid, target, prefix, countdown, limit ); }, 500 );
        }

        async showGroupStatus( rule, grp, $container, cstate, pgrp, firstGroup ) {
            var $grpel = $( '\
<div class="reactorgroup"> \
    <div class="row grouptitle"> \
        <div class="col-auto condind bg-light"><i class="bi bi-square text-muted"></i></div> \
        <div class="col"><span class="re-title">??</span></div> \
        <div class="col"><span class="currentvalue">??</span></div> \
    </div> \
    <div class="grpbody"> \
        <div class="grpcond"></div> \
    </div>\
</div>' );

            var name;
            var root = false;
            if ( "trig" === grp.id || grp.id.endsWith( '-trig' ) ) {
                name = _T(['#ruleedit-triggers-title','Triggers']);
                root = true;
            } else if ( grp.id.match( /^const?$/ ) || grp.id.endsWith( '-const' ) ) {
                name = _T(['#ruleedit-constraints-title','Constraints']);
                root = true;
            } else {
                name = grp.name ? grp.name : _T('Group {0}');
            }
            var grp_op_word = _T(['#cond-group-op-' + ( grp.op || "and" ), grp.op || "and"]);

            var title = "";
            if ( pgrp && ! firstGroup ) {
                title = _T(['#cond-group-op-' + ( pgrp.op || "and" ), pgrp.op || "and"])
                    .toLocaleUpperCase( getLocale() ) + ' ';
            }
            title += name + ' ';
            title += await RuleEditor.makeCondOptionDesc( rule, grp );
            title += ' [' +
                ( grp.invert ? `${_T(['#cond-group-op-not', 'NOT']).toLocaleLowerCase( getLocale() )} ` : "" ) +
                grp_op_word.toLocaleLowerCase( getLocale() ) + ']' +
                ( grp.disabled ? ` (${_T(['#cond-group-state-disabled','DISABLED'])})` : "" );
            // $( '.condbtn', $grpel ).text( (grp.invert ? "NOT " : "") + (grp.op || "and" ).toUpperCase() ); // ???
            $( 'span.re-title', $grpel ).text( title );
            $( 'span.currentvalue', $grpel ).text( _T(['#group-status-no-data','(no data)']) ); /* temporarily */

            $( '.row', $grpel ).attr( 'id', rule.id + '-cond-' + grp.id ); /* for status updates */

            /* Highlight groups that are "true" */
            $grpel.toggleClass( 'groupdisabled', !!grp.disabled );
            if ( !grp.disabled ) {
                var gs = cstate[ grp.id ] || {};
                if ( "undefined" === typeof gs.evalstate || null === gs.evalstate ) {
                    $grpel.addClass( "nostate" );
                } else if ( gs.evalstate ) {
                    $grpel.addClass( "truestate" );
                }
            }
            $container.append( $grpel );

            $grpel = $( 'div.grpcond', $grpel );
            var l = grp.conditions ? grp.conditions.length : 0;
            var first = true;
            for ( var i=0; i<l; i++ ) {
                var cond = grp.conditions[i];

                if ( "group" === ( cond.type || "group" ) ) {
                    this.showGroupStatus( rule, cond, $grpel, cstate, grp, i === 0 );
                } else {
                    var $row = $('<div class="row cond"></div>').attr( 'id', rule.id + '-cond-' + cond.id ); /* for status updates */

                    var condType = condTypeName[ cond.type ] !== undefined ? condTypeName[ cond.type ] : cond.type;
                    var condDesc = await RuleEditor.makeConditionDescription( rule, cond );

                    /* Operator/condition state */
                    var $el = $( '<div class="col-auto condind"><i class="bi bi-square text-muted"></i></div>' )
                        .appendTo( $row );

                    /* Condition description. Apply options to condition description */
                    if ( cond.type !== "comment" ) {
                        if ( ! first ) {
                            condType = grp_op_word.toLocaleUpperCase( getLocale() ) + " " + condType;
                        }
                        first = false;
                    }
                    let copt = await RuleEditor.makeCondOptionDesc( rule, cond );
                    if ( "" !== copt ) {
                        condDesc += '; ' + copt;
                    }
                    $row.append( $( '<div class="col"></div>' ).text( condType + ': ' + condDesc ) );

                    /* Append current value and condition state */
                    $el = $( `<div class="col"><span class="currentvalue">${_T(['#group-status-no-data','(no data)'])}</span></div>` )
                        .data( 'cond-type', cond.type ).attr( 'data-cond-type', cond.type )
                        .appendTo( $row );
                    $row.append( $el );

                    /* Apply highlight for state */
                    var cs = cstate[ cond.id ] || {};
                    if ( true ) { // cond.type !== "comment" && undefined !== currentValue ) {
                        $row.toggleClass( 'truestate', true === cs.evalstate )
                            .toggleClass( 'falsestate', false === cs.evalstate )
                            .toggleClass( 'nullstate', null === cs.evalstate || "boolean" !== cs.evalstate );
                        $( 'div.condind i', $row )
                            .toggleClass( "bi-check-square-fill text-success", true === cs.evalstate )
                            .toggleClass( "bi-x-square text-danger", false === cs.evalstate )
                            .toggleClass( "bi-square text-muted", null === cs.evalstate || "boolean" !== cs.evalstate );
                    }

                    $grpel.append( $row );
                }
            }
        }

        showExprStatus( rule, $ct, st ) {
            let exp = rule.expressions;
            let elist = Object.values( exp );
            if ( 0 === elist.length ) {
                return;
            }
            elist.sort( ( a, b ) => {
                let n1 = "undefined" === typeof a.index ? 32767 : a.index;
                let n2 = "undefined" === typeof b.index ? 32767 : b.index;
                if ( n1 === n2 ) {
                    return 0;
                }
                return n1 < n2 ? -1 : 1;
            });

            var $ex = $( '<div class="expr-status mb-2"></div>' ).appendTo( $ct );
            $( `<div class="row expr-status-title"><div class="col">${_T(['#ruleedit-expressions-title','Local Expressions'])}</div></div>` )
                .appendTo( $ex );

            elist.forEach( expr => {
                let $row = $( '<div class="row expr-status-expr"></div>' )
                    .attr( 'id', rule.id + '-expr-' + expr.name )
                    .data( 'expr', expr.name ).attr( 'data-name', expr.name )
                    .appendTo( $ex );
                let val = "";
                let stamp = "-";
                if ( ! ( st.expr || {} )[ expr.name ] || "undefined" === typeof st.expr[ expr.name ].lastvalue ) {
                    val = "" === ( expr.expr || "" ) ? _T('#expr-status-not-set') : _T('#expr-status-not-eval');
                } else {
                    let es = st.expr[ expr.name ];
                    let lastval = es.lastvalue;
                    if ( ( es.laststamp || 0 ) > 0 ) {
                        stamp = _T('since {0}', shortTime( es.laststamp ), es.laststamp );
                        $row.data( 'since', es.laststamp );
                    }
                    if ( null === lastval ) {
                        val = "(" + _T( "null" ) + ")";
                    } else {
                        let typ = Array.isArray( lastval ) ? 'array-len' : typeof lastval;
                        val = "(" +
                            _T(['#data-type-' + typ, typ], Array.isArray( lastval ) ? lastval.length : 0 ) +
                            ") ";
                        try {
                            val += JSON.stringify( lastval );
                        } catch ( err ) {
                            console.log( "Can't JSON.stringify",lastval,err);
                            val += String( lastval );
                        }
                    }
                }
                $( '<div class="col-auto expr-name"></div>' ).text( expr.name ).appendTo( $row );
                $( '<div class="col expr-expr user-select-all"></div>' ).text( expr.expr || "" ).appendTo( $row );
                $( '<div class="col expr-value user-select-all"></div>' ).text( val ).appendTo( $row );
                $( '<div class="col expr-stamp"></div>' ).text( stamp ).appendTo( $row );
            });
        }

        async handleRuleStatus( event ) {
            var $row = $( event.currentTarget ).closest( '.re-rule' );
            var rule_id = $row.attr( 'id' ).replace( /^rule-/, "" );
            var $card = $( 'div.card', $row );
            var $col;
            if ( 0 === $card.length ) {
                $card = $( '<div class="card card-body bg-transparent mx-0 my-0"></div>' );
                $( '<div class="w-100"></div>' ).appendTo( $row );
                $col = $( '<div class="col-xs-12 col-sm-12 collapse"></div>' ).appendTo( $row );
                $card.appendTo( $col );
            }
            $col = $card.closest( ".collapse" );
            if ( $col.hasClass( "show" ) ) {
                $( 'div.re-rulename i', $row ).removeClass( 'bi-chevron-down' ).addClass( 'bi-chevron-right' );
                $col.slideUp( 250, function() { $(this).removeClass( "show" ); $card.empty(); } );
            } else {
                let rule = await Rule.getInstance( rule_id );
                let cstate = await rule.getStates();
                let $ct = $( '<div class="mb-2"></div>' ).appendTo( $card.empty() );
                this.showGroupStatus( rule, rule.triggers, $ct, cstate, undefined, true );
                if ( rule.constraints && ( rule.constraints.conditions || [] ).length > 0 ) {
                    $ct = $( '<div class="mb-2"></div>' ).appendTo( $card );
                    this.showGroupStatus( rule, rule.constraints, $ct, cstate, undefined, true );
                }
                this.showExprStatus( rule, $card, cstate );
                $( '<div class="text-muted user-select-all"></div>' ).text( _T('Rule ID: {0}', rule.getID()) )
                    .appendTo( $card );
                var self = this;
                $( 'div.re-rulename i', $row ).removeClass( 'bi-chevron-right' ).addClass( 'bi-chevron-down' );
                $col.slideDown( 250, function() {
                    $(this).addClass( "show" );
                    self.showRuleStateData( rule.id, $row, cstate, true );
                } );
            }
        }

        async handleRulesetClick( ev, ui ) {
            var $navlink = $( ev.currentTarget );
            var rsid = $navlink.attr( 'id' );
            if ( "edit" === this.$tab.data( "mode" ) ) {
                if ( this.configModified ) {
                    Common.showSysModal( {
                        title: _T('Operation in Progress'),
                        body: _T('Please complete the current operation before switching views'),
                    } );
                    console.log( "Ignoring ruleset click while in editor" );
                    return;
                }
                /* Close editors */
                this.close_editor( rsid );
                return false;
            }
            $( 'div#rulesets .list-group-item' ).removeClass( 'ui-selected' );
            $( 'li#nav-item-rules ul#ruleset-nav > li.nav-item a.nav-link' ).removeClass( 'active' );
            $navlink.addClass( 'active' );
            localStorage.setItem( 'last_ruleset', rsid );

            var $container = $( 'div#rules' ).empty();
            $( 'div#rightheader span#rulesetname' ).text( _T("Loading...") );

            var set = await Ruleset.getInstance( rsid );
            var rules = await set.getRules();
            rules.forEach( rule => {
                var $row = $( '<div class="re-rule row align-items-center"></div>' )
                    .attr( 'id', 'rule-' + rule.id )
                    .appendTo( $container );
                $( '<div class="col re-rulename"></div>' )
                    .text( rule.name || rule.id )
                    .append( '<i class="bi bi-chevron-right ms-1"></i>' )
                    .appendTo( $row );
                $( '<div class="col re-rulestatus"></div>' )
                    .appendTo( $row );
                $( `<div class="col text-end text-nowrap re-rulecontrols">
  <button class="btn bi-btn ms-1 ruleEnable" title="${_T(['#rule-control-enable','Toggle Enabled'])}"><i class="bi"></i></button>
  <button class="btn bi-btn ms-1 ruleEdit text-primary" title="${_T(['#rule-control-edit','Edit Rule'])}"><i class="bi bi-pencil"></i></button>
  <button class="btn bi-btn ms-1 ruleReset text-warning" title="${_T(['#rule-control-reset','Reset Rule'])}"><i class="bi bi-bootstrap-reboot"></i></button>
  <button class="btn bi-btn ms-1 ruleCopy text-secondary" title="${_T(['#rule-control-copy','Copy/Move'])}"><i class="bi bi-share"></i></button>
  <button class="btn bi-btn ms-1 ruleDrag text-secondary draghandle" title="${_T(['#rule-control-sort','Sort (click and drag)'])}"><i class="bi bi-arrow-down-up"></i></button>
  <button class="btn bi-btn ms-4 ruleDelete text-danger" title="${_T(['#rule-control-delete','Delete Rule'])}"><i class="bi bi-x"></i></button>
</div>` ).appendTo( $row );
                $row.toggleClass( 'rule-disabled', !!rule.triggers.disabled );
                $( 'button.ruleEnable i', $row )
                    .toggleClass( 'bi-toggle-off', !!rule.triggers.disabled )
                    .toggleClass( 'bi-toggle-on', !rule.triggers.disabled )
                    .on( 'click.reactor', this.handleRuleEnableToggle.bind( this ) );
                $( 'button.ruleEdit', $row ).on( 'click.reactor', this.edit.bind( this ) );
                $( 'button.ruleReset', $row ).on( 'click.reactor', this.handleRuleReset.bind( this ) );
                $( 'button.ruleCopy', $row ).on( 'click.reactor', this.handleRuleCopy.bind( this ) );
                $( 'button.ruleDelete', $row ).on( 'click.reactor', this.handleRuleDelete.bind( this ) );

                /* Set up state display */
                let self = this;
                rule.getStatesDataObject().then( dobj => {
                    api.subscribe( rule.getDataObject(), function( msg ) {
                        if ( "data-changed" === msg.type ) {
                            self.showRuleStateData( rule.id, $row, dobj.value );
                        } else if ( "data-deleted" === msg.type ) {
                            console.log("**** DELETED DATA ****", msg );
                        }
                    });
                    api.subscribe( dobj, self.handleDataObjectChange.bind( this ) );
                    self.showRuleStateData( rule.id, $row, dobj.value, true );
                });
            });

            if ( 0 === rules.length ) {
                $( `<span id="ruleset-empty">${_T('This rule set is empty. Click "Add Rule" to add a rule to it.')}</span>` )
                    .appendTo( $container );
            }

            $container.data( 'ruleset', set.id );
            $( 'div#rightheader span#rulesetname' ).text( set.name || set.id );
            $( 'div#rightheader button#deleteruleset' ).prop( 'disabled', set.rules.length > 0 );

            $( '.re-rulename', $container ).on( 'click.reactor', this.handleRuleStatus.bind( this ) );
        }

        async handleRulesetRenameEvent( ev ) {
            var rsid = $( 'div#rules' ).data( 'ruleset' );
            var set = await Ruleset.getInstance( rsid );
            var $form = $( '<div class="form-group"><input id="newname" class="form-control form-control-sm"></div>' );
            $( 'input', $form ).val( set.name || set.id );
            showSysModal({
                title: _T(['#ruleset-rename-title','Rename Rule Set']),
                body: $form,
                buttons: [{
                    label: _T(['#ruleset-rename-button-cancel','Cancel']),
                    close: true,
                    class: "btn-primary"
                },{
                    label: _T(['#ruleset-rename-button-rename','Rename']),
                    event: "rename",
                    class: "btn-success"
                }
            ]
            }).then( data => {
                if ( "rename" === data ) {
                    var newname = $( '#sysdialog #newname' ).val() || "";
                    if ( "" !== newname && newname !== set.name ) {
                        set.getDataObject().value.name = newname;
                        set.getDataObject().save().then( () => {
                            $( 'div#rightheader span#rulesetname' ).text( newname );
                            $( 'ul#ruleset-nav a#' + idSelector( set.getID() ) + '.nav-link' )
                                .text( newname );
                        });
                    }
                }
            });
        }

        async show_list( ruleset_id ) {
            let self = this;

            this.$tab.empty();
            this.$tab.data( 'mode', 'list' ).attr( 'data-mode', 'list' );

            /* Before we empty the subnavigation, save the active element, if any */
            var $subnav = $( 'li#nav-item-rules ul#ruleset-nav' );
            var lastActiveRuleset = ruleset_id || $( 'li.nav-item a.nav-link.active', $subnav ).attr( 'id' );
            var redraw = $subnav.hasClass( "show" );
            /** After possible hide, if the previous display had no active item or wasn't displayed
             *  at all, see if local storage has a previous item. */
            if ( ! lastActiveRuleset ) {
                lastActiveRuleset = localStorage.getItem( 'last_ruleset' );
            }
            $subnav.empty();

            $( `<div id="rightheader">
  <span id="rulesetname" class="mx-2" title="${_T(['#ruleset-control-rename','Click to rename'])}"></span>
  <div class="d-inline">
    <button id="sortalpha" class="bi-btn bi-btn-white" title="${_T('#ruleset-control-alphasort')}"><i class="bi bi-sort-alpha-down"></i></button>
    <button id="deleteruleset" class="bi-btn bi-btn-white" title="${_T('#ruleset-control-delete')}"><i class="bi bi-x"></i></button>
  </div>
  <div class="float-end mt-1">
    <button id="addrule" class="btn btn-sm btn-success">${_T(['#ruleset-button-newrule','Create Rule'])}</button>
  </div>
</div>` )
                .appendTo( this.$tab );
            var $rulelist = $( '<div id="rules" class="container-fluid"></div>' )
                .appendTo( this.$tab );
            // $rulelist.css( 'max-height', Math.floor( ( window.innerHeight - $rulelist.offset().top ) ) + "px" );

            var rulesets = await Rulesets.getRulesets();
            rulesets.forEach( set => {
                if ( null === set ) { /* unable to fetch set in sequence */
                    return;
                }

                var $nav = $( '<li class="nav-item re-nav-sub w-100"></li>' )
                    .appendTo( $subnav );
                $( '<a href="#" class="nav-link "></a>' )
                    .attr( 'id', set.id )
                    .text( set.name || set.id )
                    .on( 'click.reactor', this.handleRulesetClick.bind( this ) )
                    .appendTo( $nav );
                // $( '<span class="badge bg-info rounded-pill ms-2 align-top"></span>' ).text( (set.rules || []).length ).appendTo( $nav );
            });

            if ( !redraw ) {
                $subnav.removeClass( "hide" ).addClass( "show" ).slideDown();
            }

            $( 'li#nav-item-rules div.nav-item-controls' ).show();

            /* The ruleset list itself is sortable. */
            $subnav.sortable({
                items: "> .nav-item",
                axis: "y",
                update: this.handleRulesetSort.bind( this ),
                placeholder: "sort-placeholder",
                distance: 15
            });

            /* Rules are also sortable */
            $rulelist.sortable({
                items: "> .re-rule",
                axis: "y",
                handle: 'button.ruleDrag',
                cancel: '', /* so draghandle can be button */
                containment: $rulelist,
                update: this.handleRuleSort.bind( this ),
                placeholder: "sort-placeholder"
            });

            $( 'button#sortalpha' ).on( 'click.reactor', async function() {
                var ruleset_id = $( 'div#rules' ).data( 'ruleset' );
                var ruleset = await Ruleset.getInstance( ruleset_id );
                var rules = await ruleset.getRules();
                rules.sort( function( a,b ) {
                    return String(a.name || a.id).localeCompare( b.name || b.id, undefined, { sensitivity: 'base' } );
                });
                var dobj = ruleset.getDataObject();
                dobj.value.rules = rules.map( el => el.id );
                dobj.save().then( () => {
                    /* Display new sort order */
                    var $insert = $( 'div#rules .re-rule:last' );
                    rules.forEach( rule => {
                        var rid = rule.getID();
                        var $el = $( 'div#rules div[id="rule-' + rid + '"].re-rule' );
                        if ( $el.attr( 'id' ) !== $insert.attr( 'id' ) ) {
                            $el.detach().insertAfter( $insert );
                        }
                        $insert = $el;
                    });
                });
            });

            $( 'button#addrule', this.$tab ).on( 'click.reactor', async function() {
                var rsid = $( 'div#rules' ).data( 'ruleset' );
                var ruleset = await Ruleset.getInstance( rsid );
                var rule = await Rule.create( ruleset );
                self.editRule( rule, true );
            });

            $( 'span#rulesetname', this.$tab ).on( 'click.reactor', this.handleRulesetRenameEvent.bind( this ) );

            $( 'button#deleteruleset', this.$tab ).on( 'click.reactor', async function() {
                var rsid = $( 'div#rules', self.$tab ).data( 'ruleset' );
                let set = await Ruleset.getInstance( rsid );
                if ( set.rules.length > 0 ) {
                    showSysModal( { title: "Unable to Delete",
                        body: "The rule set cannot be deleted because it is not empty." } );
                    return;
                }
                var $nav = $( 'ul#ruleset-nav' );
                var $row = $( 'li.nav-item a[id="' + idSelector( rsid ) + '"].nav-link', $nav ).closest( 'li.nav-item' );
                /* Find next row to show; if we deleted last displayed, no next, so get prev. */
                var $nextrow = $row.next();
                if ( 0 === $nextrow.length ) {
                    $nextrow = $row.prev();
                }
                /* If not prev either, we're trying to delete last ruleset. Verboten! */
                if ( 0 === $nextrow.length ) {
                    showSysModal( {
                        title: _T(['#ruleset-delete-unable','Unable to Delete Rule Set']),
                        body: _T(['#ruleset-delete-message',"You cannot delete the last rule set; there must be at least one."])
                    } );
                    return;
                }
                set.delete().then( () => {
                    $row.remove();
                    $( 'a.nav-link', $nextrow ).click();
                });
            });

            /* Finally, select a ruleset to display. If we have an active item, try to restore. */
            var $i = $( 'li.nav-item a#' + idSelector( lastActiveRuleset || "_" ) + '.nav-link', $subnav );
            if ( 1 === $i.length ) {
                $i.click();
            } else {
                /* Default select first one */
                $( 'li:first a.nav-link', $subnav ).click();
            }

        }

        close_editor( ruleset_id ) {
            Object.values( this.editors ).forEach( function( ed ) { ed.close(); });
            this.$tab.empty().data( "rule", "" ).attr( "data-rule", "" )
                .data( "mode", "list" ).attr( "data-mode", "list" );
            this.editrule = false;
            this.newrule = false;
            this.configModified = false;
            this.show_list( ruleset_id );
        }

        /**
         * Handle save/revert click
         */
        async handleSaveRevertClick( event ) {
            var $el = $( event.target );
            $( '.saveconf,.revertconf', this.$tab ).prop( 'disabled', true );
            var rule_id = this.$tab.data( 'rule' );
            var rule = await Rule.getInstance( rule_id );
            if ( $el.hasClass('revertconf') ) {
                if ( this.configModified ) {
                    showSysModal( {
                        title: _T(['#ruleedit-unsaved-title','Rule Modified (Unsaved Changes)']),
                        body: _T(['#ruleedit-unsaved-prompt','You have unsaved changes to this rule. Really discard changes and exit?']),
                        buttons: [{
                            label: _T(['#ruleedit-unsaved-button-discard',"Discard Changes"]),
                            event: "discard",
                            class: "btn-danger"
                        },{
                            label: _T(['#ruleedit-unsaved-button-cancel',"Cancel"]),
                            close: true,
                            class: "btn-success"
                        }]
                    }).then( async event => {
                        if ( "discard" === event ) {
                            this.configModified = false;
                            if ( this.newrule ) {
                                await rule.delete();
                            }
                            this.close_editor();
                            return;
                        }
                        this.updateSaveControls();
                    });
                } else {
                    if ( this.newrule ) {
                        await rule.delete();
                    }
                    this.close_editor();
                }
                return false;
            } else if ( $el.hasClass( 'saveconf' ) ) {
                var doexit = $el.hasClass( 'saveexit' );

                console.log("Saving rule ", rule_id, ", editrule", this.editrule);

                var dobj = rule.getDataObject();
                dobj.value.name = this.editrule.name;
                dobj.value.serial = ( dobj.value.serial || 0 ) + 1;

                /* Grab the data from the embedded editors */
                dobj.value.triggers = this.editors.trig_editor.current();
                dobj.value.triggers.id = 'trig';
                dobj.value.constraints = this.editors.const_editor.current();
                dobj.value.constraints.id = 'cons';
                dobj.value.react_set = this.editors.set_editor.current();
                dobj.value.react_reset = this.editors.reset_editor.current();
                dobj.value.expressions = this.editors.expr_editor.current();
                dobj.save().then( dobj => {
                    this.configModified = false;
                    this.newrule = false;
                    Object.values( this.editors ).forEach( function( ed ) {
                        ed.notifySaved();
                    });
                    if ( doexit ) {
                        this.close_editor();
                    } else {
                        this.updateSaveControls();
                    }
                }).catch( function( e ) {
                    console.log(this,"save failed",e);
                    showSysModal({
                        title: _T('#dialog-error-title'),
                        body: _T('The save failed; try again. {0}', e),
                    }).then( function( data ) {
                        this.updateSaveControls();
                    });
                });

                return false;
            }
        }

        async edit( event ) {
            var $row = $( event.currentTarget ).closest( '.re-rule' );
            var rule_id = $row.attr( 'id' ).replace( /^rule-/, "" );

            var rule = await Rule.getInstance( rule_id );
            if ( !rule ) {
                return;
            }

            this.editRule( rule );
        }
    }

    var tabInstance = false;

    return {
        "init": function( $main ) {
            if ( ! tabInstance ) {
                tabInstance = new RulesTab( $main );
            }
        },
        "tab": () => tabInstance.getTabElement()
    };

})(jQuery);
